package uk.ac.starlink.util.images;

/**
 * Simple class for placeholding a directory of images. 
 *
 * Use this to access images using the getResource method of Class. 
 * For instance the code snippet following loads the "exit.gif" image.
 * <pre>
 *    import uk.ac.starlink.util.images.ImageHolder;
 *    ....
 *    ImageIcon icon = ImageHolder.class.getResource( "exit.gif" );
 * </pre>
 */
public class ImageHolder 
{
    //  Does nothing.
}
